-- The control unit for the UART receiver 
-- A finite state machine which will generate the output signals
-- and will command the datapath depending on the externals inputs 
-- and the status signals from the datapath 

LIBRARY ieee;
USE  IEEE.STD_LOGIC_1164.all;
USE  IEEE.STD_LOGIC_ARITH.all;
USE  IEEE.STD_LOGIC_UNSIGNED.all;

			
ENTITY receiver_control_unit IS
      Port (	
		Start_Bit   	: IN	STD_LOGIC;
		CLK 			: IN	STD_LOGIC;
		CD 		        : IN	STD_LOGIC;
		Enable_RX 	    : IN	STD_LOGIC;
		EP_check  	    : IN 	STD_LOGIC;	
		Stop_Bit   		: IN 	STD_LOGIC;
		TC4,TC8,TC10	: IN	STD_LOGIC;
		S 	            : OUT 	STD_LOGIC_VECTOR (1 DOWNTO 0);
		LDD  	        : OUT	STD_LOGIC;
		CE_MOD4  	    : OUT 	STD_LOGIC;
		CE_MOD8     	: OUT 	STD_LOGIC;
		Clear_Counters	: OUT 	STD_LOGIC;
		Error    		: OUT 	STD_LOGIC;
		New_Data   		: OUT 	STD_LOGIC
	);

END receiver_control_unit;

			-- Internal desciption in FSM style

ARCHITECTURE FSM_like OF receiver_control_unit IS

-- Internal wires 
				-- State signals declaration 
				-- A BCD counter will consist of 10 diferent states
	TYPE State_type IS (Idle, wait_4cycles, wait_8cycles, sample_serial_in, test_parity,
	                     test_stop_bit, error_signal, load_data_reg, data_ready) ;  
--------> This is important: specifying to the synthesiser the code for the states
	ATTRIBUTE syn_encoding : STRING;
	ATTRIBUTE syn_encoding OF State_type : TYPE IS "sequential";
		-- (It may be:"sequential" (binary); "gray"; "one-hot", etc.
	SIGNAL present_state, future_state	: State_type ; 

-- Constants for special states (the first and the last state)
CONSTANT Reset		: State_type := Idle; 	-- The fisrt state. This is another name for the state Idle
											-- so that the state registes doesn't have to be modified 

BEGIN
---------------------- State Register: the only clocked block.  
---------------------- The "memory" of the system (future events will depend on past events 
state_register: PROCESS (CD,CLK, future_state)
	BEGIN
						
	IF CD = '1' THEN 			-- reset counter
		present_state <= Reset; 
	ELSIF (CLK'EVENT and CLK = '1') THEN	-- Synchronous register (D-type flip-flop)
		present_state <= future_state; 
	END IF;

END PROCESS state_register;

------------------------- CC1: Combinational system for calculating next state

CC_1: PROCESS (present_state, Stop_bit, EP_check, TC10, TC8, TC4,Enable_RX,Start_bit)
	BEGIN	
			CASE present_state IS
				WHEN Idle =>
					IF (Start_bit = '0' and Enable_RX = '1') THEN
					   future_state <= wait_4cycles;
					ELSE 
					   future_state <= Idle  ;
					END IF;
					
				WHEN wait_4cycles => 
				    IF (TC4 = '1' and Start_bit = '0') THEN
				       future_state <= wait_8cycles;
				    ELSIF (TC4 = '1' and Start_bit = '1')THEN
				       future_state <= Idle;
				    ELSE
					   future_state <= wait_4cycles;
					END IF;
					
				WHEN wait_8cycles => 
				    IF (TC8 = '1') THEN
				       future_state <= sample_serial_in;
				    ELSE 
				       future_state <= wait_8cycles;
					END IF;
					
				WHEN sample_serial_in => 
				    IF (TC10 = '1') THEN 
						future_state <= test_parity;
					ELSE
						future_state <= wait_8cycles;
					END IF;
					 
				WHEN test_parity => 
				    IF (EP_check = '1') THEN
				      future_state <= error_signal;
				    ELSE 
				      future_state <= test_stop_bit;
					END IF;
					
				WHEN error_signal => 
				      future_state <= Idle;
					
				WHEN test_stop_bit => 
				    IF (Stop_Bit = '0') THEN
					   future_state <= error_signal;
					ELSE 
					   future_state <= load_data_reg;
					END IF;
					
				WHEN load_data_reg =>
					future_state <= data_ready ;
					
				WHEN data_ready => 
					future_state <= Idle ;
									
			END CASE ;		
END PROCESS CC_1;

------------------------- CS_2: combinational system for calculating extra outputs 
-------------------------- and outputing the present state (the actual count)

CC_2: PROCESS (present_state)
	BEGIN
		CASE present_state IS
				WHEN Idle => 
					S 	            <= "00";
					LDD  	        <= '0';
					CE_MOD4  	    <= '0';
					CE_MOD8     	<= '0';
					Clear_Counters	<= '0';
					Error    		<= '0';
					New_Data   		<= '0';
					
				WHEN wait_4cycles => 
				    S 	            <= "00";
					LDD  	        <= '0';
					CE_MOD4  	    <= '1';
					CE_MOD8     	<= '0';
					Clear_Counters	<= '0';
					Error    		<= '0';
					New_Data   		<= '0';
					
				WHEN wait_8cycles => 
				    S 	            <= "00";
					LDD  	        <= '0';
					CE_MOD4  	    <= '0';
					CE_MOD8     	<= '1';
					Clear_Counters	<= '0';
					Error    		<= '0';
					New_Data   		<= '0';
					
				WHEN sample_serial_in => 
				    S 	            <= "01";
					LDD  	        <= '0';
					CE_MOD4  	    <= '0';
					CE_MOD8     	<= '1';
					Clear_Counters	<= '0';
					Error    		<= '0';
					New_Data   		<= '0';
					 
				WHEN test_parity => 
				    S 	            <= "00";
					LDD  	        <= '0';
					CE_MOD4  	    <= '0';
					CE_MOD8     	<= '0';
					Clear_Counters	<= '0';
					Error    		<= '0';
					New_Data   		<= '0';
					
				WHEN error_signal => 
				    S 	            <= "00";
					LDD  	        <= '0';
					CE_MOD4  	    <= '0';
					CE_MOD8     	<= '0';
					Clear_Counters	<= '1';
					Error    		<= '1';
					New_Data   		<= '0';
					
				WHEN test_stop_bit => 
				    S 	            <= "00";
					LDD  	        <= '0';
					CE_MOD4  	    <= '0';
					CE_MOD8     	<= '0';
					Clear_Counters	<= '0';
					Error    		<= '0';
					New_Data   		<= '0';
					
				WHEN load_data_reg =>
					S 	            <= "00";
					LDD  	        <= '1';
					CE_MOD4  	    <= '0';
					CE_MOD8     	<= '0';
					Clear_Counters	<= '0';
					Error    		<= '0';
					New_Data   		<= '0';
					
				WHEN data_ready => 
					S 	            <= "00";
					LDD  	        <= '0';
					CE_MOD4  	    <= '0';
					CE_MOD8     	<= '0';
					Clear_Counters	<= '1';
					Error    		<= '0';
					New_Data   		<= '1';
									
			END CASE ;	

END PROCESS CC_2;

-- Place other logic if necessary

END FSM_like;






